from pyramid.httpexceptions import HTTPOk
from pyramid.settings import asbool
from sentry_sdk import capture_message


class DataTablesAPI(object):
    """DataTables API: return empty list if the request isn't made from dataTables."""

    def __init__(self, request):
        self.request = request

        if not asbool(self.request.GET.get('datatables')):
            capture_message('Invalid API call.')
            # Raise HTTPOk here so that we don't enter the view function.
            raise HTTPOk(json=[])


def manage_api_queries(query_string):
    """Manage "classic" API query."""
    limit = query_string.get('limit')
    offset = query_string.get('offset')
    if (limit and not limit.isdigit()) or (offset and not offset.isdigit()):
        raise KeyError()

    sort = [q for q in query_string.get('_sort').split(',')] if query_string.get('_sort') else []
    full_text_search = query_string.get('_q')

    # Manage searches with multiple values (?key=value1&key=value2).
    search = []
    for key, value in query_string.mixed().items():
        if key not in ('limit', 'offset', '_sort', '_q'):
            if isinstance(value, list):
                search.append((key, 'in', value))
            else:
                search.append((key, '==', value))

    return {
        'full_text_search': full_text_search,
        'limit': limit,
        'offset': offset,
        'search': search,
        'sort': sort,
    }


def manage_datatables_queries(query_string):
    """Manage datatables queries.
    Draw is a datables variable that needs to be returned in the response.
    """
    draw = int(query_string.get('draw'))
    offset = query_string.get('start')
    limit = query_string.get('length')

    if not draw or not offset or not limit:
        raise KeyError()

    full_text_search = query_string.get('search[value]')
    # Prevent people pasting a full novel in the search input from killing the server with a crazy SQL search.
    if full_text_search and len(full_text_search.split(' ')) > 5:
        full_text_search = None

    search = []
    search_qs = query_string.get('filter')
    if search_qs:
        search_qs_split = search_qs.split('&&')
        for criteria in search_qs_split:
            if '==' in criteria:
                params = criteria.split('==')
                search.append((params[0], '==', params[1]))
            elif '!=' in search_qs:
                params = criteria.split('!=')
                search.append((params[0], '!=', params[1]))
            else:
                raise KeyError()

    # When filtering on a given column (columns[column][search][value]), Datatables will use commas-concatenated
    # strings for multiple values. Split those values and add them to search parameters.
    i = 0
    while query_string.get(f'columns[{str(i)}][search][value]') is not None:
        name = query_string.get(f'columns[{str(i)}][data]')
        value = str(query_string.get(f'columns[{str(i)}][search][value]'))
        if value:
            value_list = value.split(',')
            if len(value_list) > 1:
                search.append((name, 'in', value_list))
            else:
                search.append((name, '==', value))
        i += 1

    # To define the query order, Datatables sends order[sort_order][column] = column_number and
    # order[sort_order][dir] = 'asc' or 'desc'. So we loop through the different keys and format the data according
    # to our needs. (Attribute names are sent in columns[column_number][data]).
    sort = []
    i = 0
    while query_string.get(f'order[{str(i)}][column]'):
        column = str(query_string.get(f'order[{str(i)}][column]'))
        attribute = query_string.get(f'columns[{column}][data]')
        direction = '-' if query_string.get(f'order[{str(i)}][dir]') == 'desc' else ''
        sort.append(f'{direction}{attribute}')
        i += 1

    return {
        'draw': draw,
        'full_text_search': full_text_search,
        'limit': limit,
        'offset': offset,
        'search': search,
        'sort': sort,
    }
