"""Authentication and authorization.
Callbacks:
    - get user (validate unauthenticated id)
    - get user locale
    - get user principals (rights)

Authentication:
    - OpenID Connect policy
    - cleaned callback policy

Authorization: Tenanted policy.
Authorization utilities: get rights without tenants.
"""
import binascii
from collections import namedtuple
from datetime import datetime, timedelta

from pyramid.authentication import b64decode, CallbackAuthenticationPolicy
from pyramid.compat import is_nonstr_iter
from pyramid.httpexceptions import HTTPBadRequest, HTTPServiceUnavailable, HTTPUnauthorized
from pyramid.interfaces import IAuthenticationPolicy, IAuthorizationPolicy
from pyramid.location import lineage
from pyramid.security import ACLAllowed, ACLDenied, Allow, Authenticated, Everyone
from zope.interface import implementer

from parsys_utilities import AVAILABLE_LOCALES, SERVICE_SESSION_DURATION, USER_SESSION_DURATION
from parsys_utilities.openid_client import OpenID, request_user_info

HTTPBasicCredentials = namedtuple('HTTPBasicCredentials', ['username', 'password'])

Right = namedtuple('Right', ['name', 'tenant'])


def authenticate_rta(request):
    """Authenticate RTA using "reversed" authentication: the client has credentials to authenticate on RTA
    (client_id/secret). Make RTA use those same credentials for the client API.

    Args:
        request (pyramid.request.Request).
    """
    rta_auth = extract_http_basic_credentials(request)
    if not rta_auth:
        return False

    client_id = request.registry.settings['rta.client_id']
    secret = request.registry.settings['rta.secret']

    return client_id == rta_auth.username and secret == rta_auth.password


class CleanedCallbackAuthenticationPolicy(CallbackAuthenticationPolicy):
    """Simple patch on pyramid's CallbackAuthenticationPolicy, to exclude rights using reserved names."""

    # noinspection PyMethodMayBeStatic
    def _clean_principal(self, princid):
        """Remove reserved rights (Authenticated, Everyone, g:admin and userid:...).

        Args:
            princid (str): Right name.

        Returns:
            str: princid or None if right was reserved.
        """
        if princid in (Authenticated, Everyone, 'g:admin') or princid.startswith('userid:'):
            princid = None
        return princid

    def effective_principals(self, request):
        """Copy/Paste of parent method except for prefixing userid. Refer to Pyramid for the full doc.

        As Pyramid adds the user_id to the rights lists, prevent a stupid attack where the user_id would be the name of
        a right, thus giving access to an unauthorized user.
        """
        effective_principals = [Everyone]
        # noinspection PyUnresolvedReferences
        userid = self.unauthenticated_userid(request)

        if userid is None:
            return effective_principals

        if self._clean_principal(userid) is None:
            return effective_principals

        if self.callback is None:
            groups = []
        else:
            groups = self.callback(userid, request)

        if groups is None:
            return effective_principals

        effective_principals.append(Authenticated)
        # Prefix the userid.
        effective_principals.append(f'userid:{userid}')
        effective_principals.extend(groups)

        return effective_principals


def extract_http_basic_credentials(request, header='Authorization'):
    """A helper function for extraction of HTTP Basic credentials
    from a given :term:`request`.

    Returns a :class:`.HTTPBasicCredentials` 2-tuple with ``username`` and
    ``password`` attributes or ``None`` if no credentials could be found.

    Copied from pyramid.authentication, adding the possibility to set the header's name.
    """
    authorization = request.headers.get(header)
    if not authorization:
        return None

    try:
        authmeth, auth = authorization.split(' ', 1)
    except ValueError:  # Not enough values to unpack.
        return None

    if authmeth.lower() != 'basic':
        return None

    try:
        authbytes = b64decode(auth.strip())
    except (TypeError, binascii.Error):  # Can't decode.
        return None

    # Try utf-8 first, then latin-1; see discussion in https://github.com/Pylons/pyramid/issues/898.
    try:
        auth = authbytes.decode('utf-8')
    except UnicodeDecodeError:
        auth = authbytes.decode('latin-1')

    try:
        username, password = auth.split(':', 1)
    except ValueError:  # Not enough values to unpack.
        return None

    return HTTPBasicCredentials(username, password)


# noinspection PyUnusedLocal
def get_effective_principals(userid, request, allow_admins=False):
    """Pyramid authentication policy callback, get the user's principals.
    We validate in get_user that the user info is valid, so we can serenely return user.rights here.
    Some apps accept admins, some don't.

    Args:
        userid (str).
        request (pyramid.request.Request).
        allow_admins (bool).

    Returns:
        str: the locale to use.
    """
    if not request.user:
        return

    if request.user.is_admin:
        if allow_admins:
            return ['g:admin']
    else:
        return [Right(tenant=item[0], name=item[1]) for item in request.user.rights]


def get_user(request):
    """Validate that the user info contained in the session is valid (= the access token isn't expired).
    If the access token is expired, try to get a new one using the refresh token.

    Args:
        request (pyramid.request.Request).

    Returns:
        namedtuple: user info.
    """
    user_id = request.unauthenticated_userid
    if not user_id:
        return

    access_token = request.session.get('access_token')
    # If the access token is valid, use the user info.
    if access_token and access_token['expires_at'] > datetime.utcnow():
        Principal = namedtuple('Principal', request.session['user'].keys())
        return Principal(**request.session['user'])
    # If the token is expired, delete it.
    elif access_token:
        del request.session['access_token']

    # If the access token is expired but the refresh token is valid, retrieve a new access token and use the user info.
    refresh_token = request.session.get('refresh_token')
    if refresh_token and refresh_token['expires_at'] > datetime.utcnow():
        # RTA returns the same refresh token, so there is no need to persist it.
        try:
            access_token, _refresh_token = OpenID(request).request_access_token_with_refresh_token(
                refresh_token['token']
            )
            request.session['access_token'] = access_token
            Principal = namedtuple('Principal', request.session['user'].keys())
            return Principal(**request.session['user'])
        # Invalid refresh token.
        except (HTTPBadRequest, HTTPServiceUnavailable, HTTPUnauthorized):
            del request.session['refresh_token']
    # If the token is expired, delete it.
    elif refresh_token:
        del request.session['refresh_token']

    # If both tokens are expired, forget the user.
    if 'user' in request.session:
        del request.session['user']


def get_user_locale(request):
    """Get the locale the app should be using.
    If a user is authenticated, use his locale, otherwise use the browser's headers.

    Args:
        request (pyramid.request.Request).

    Returns:
        str: the locale to use.
    """
    if request.user:
        return request.user.locale
    else:
        default = request.registry.settings.get('pyramid.default_locale_name') or 'en'
        return request.accept_language.lookup(AVAILABLE_LOCALES, default=default)


def get_tenants_for_right(request, right):
    """List tenants for which the user has a specific right.

    Args:
        request (pyramid.request.Request).
        right (str).

    Returns:
        list(dict): tenants in the form {'id': XX, 'name': XX}.
    """
    tenants = request.user.tenants
    principals = request.effective_principals
    return [tenant for tenant in tenants if (tenant['id'], right) in principals]


@implementer(IAuthenticationPolicy)
class OpenIDConnectAuthenticationPolicy(CleanedCallbackAuthenticationPolicy):
    """OpenID Connect authentication policy: use the session to persist user info/access token/refresh token.
    Use the tokens to retrieve the user info when needed.

    Constructor args:
        callback (func): pyramid auth callback, returning the user's principals.
        authorize_services (bool): are services authorized in this app?
        debug (bool).
    """

    def __init__(self, callback=None, authorize_services=False, debug=False):
        self.callback = callback
        self.authorize_services = authorize_services
        self.debug = debug

    @staticmethod
    def get_openid_client_auth(request):
        """Get the app's OpenID Connect authentication data.

        Args:
            request (pyramid.request.Request).

        Returns:
            str: app ID.
            str: app secret.
        """
        client_id = request.registry.tenant_config.get_for_domain('rta.client_id', request.host_url)
        secret = request.registry.tenant_config.get_for_domain('rta.secret', request.host_url)
        return client_id, secret

    def get_user_info_with_access_token(self, request):
        """If there is no user info in the session but there is a valid access token (case: returning from settings
        # page), request the user info.

        Args:
            request (pyramid.request.Request).

        Returns:
            str: userid or None if user isn't authenticated.
        """
        access_token = request.session.get('access_token')
        if not access_token:
            return

        # If the token is expired, delete it.
        if access_token['expires_at'] < datetime.utcnow():
            del request.session['access_token']
            return

        openid_client_auth = self.get_openid_client_auth(request)
        request.session['user'] = request_user_info(request, access_token['token'], openid_client_auth)

        return request.session['user']['id']

    def get_user_info_with_refresh_token(self, request):
        """If there is no user info in the session but there is a valid refresh token (case: returning from settings
        # page), request the user info.

        Args:
            request (pyramid.request.Request).

        Returns:
            str: userid or None if user isn't authenticated.
        """
        refresh_token = request.session.get('refresh_token')
        if not refresh_token:
            return

        # If the token is expired, delete it.
        if refresh_token['expires_at'] < datetime.utcnow():
            del request.session['refresh_token']
            return

        token = refresh_token['token']
        try:
            new_access_token, refresh_token = OpenID(request).request_access_token_with_refresh_token(token)
        # Invalid refresh token.
        except (HTTPBadRequest, HTTPServiceUnavailable, HTTPUnauthorized):
            del request.session['refresh_token']
            return

        request.session['access_token'] = new_access_token

        return self.get_user_info_with_access_token(request)

    @staticmethod
    def get_legacy_access_token(request):
        """Legacy access token, in query string.

        Args:
            request (pyramid.request.Request).

        Returns:
            str: token or None.
        """
        return request.headers.get('X-Cosmos-Auth-Token') or request.GET.get('X-COSMOS-AUTH-TOKEN')

    @staticmethod
    def get_services_access_token(request):
        """OAuth2 bearer token.

        Args:
            request (pyramid.request.Request).

        Returns:
            str: token or None.
        """
        if request.authorization and request.authorization.authtype == 'Bearer':
            return request.authorization.params

    def authenticate_service(self, request):
        """Authenticate a service. This means retrieving the bearer token, validating it with RTA, then persisting
        the user info in the session (only lasts for the duration of the request).

        Args:
            request (pyramid.request.Request).

        Returns:
            str: userid or None if user isn't authenticated.
        """
        # Retrieve the access token.
        access_token = self.get_legacy_access_token(request) or self.get_services_access_token(request)
        if not access_token:
            return

        # Validate it.
        try:
            openid_client_auth = self.get_openid_client_auth(request)
            user_info = request_user_info(request, access_token, openid_client_auth)
        except (HTTPBadRequest, HTTPServiceUnavailable, HTTPUnauthorized):
            return

        # Only services can authenticate this way.
        if user_info.get('is_service', False):
            expires_at = datetime.utcnow() + timedelta(seconds=SERVICE_SESSION_DURATION)
            access_token = {'token': access_token, 'expires_at': expires_at}
            self.remember(request, user_info['id'], user=user_info, access_token=access_token, refresh_token={})

            if hasattr(request.session, 'adjust_timeout_for_session'):
                request.session.adjust_timeout_for_session(SERVICE_SESSION_DURATION)

            return user_info['id']

    def unauthenticated_userid(self, request):
        """Get the unauthenticated_userid.
        Either:
            - get it from the session
            - if the user info is missing, but an access or a refresh token is present, try to reauthenticate the user
            - authenticate a service.

        Args:
            request (pyramid.request.Request).

        Returns:
            str: userid or None if user isn't authenticated.
        """
        try:
            return request.session['user']['id']
        except (KeyError, TypeError):
            pass

        # In the specific case where we deleted the user info but kept the access/refresh tokens (Settings page),
        # retrieve the user info using the tokens.
        user_id = self.get_user_info_with_access_token(request)
        if user_id:
            return user_id

        user_id = self.get_user_info_with_refresh_token(request)
        if user_id:
            return user_id

        # Manage services.
        if self.authorize_services:
            return self.authenticate_service(request)

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def remember(self, request, userid, user, access_token, refresh_token):
        """Remember the current user.

        Args:
            request (pyramid.request.Request).
            userid (str).
            user (dict): user info.
            access_token (dict): token/expires_at.
            refresh_token (dict): token/expires_at.
        """
        request.session['user'] = user
        request.session['access_token'] = access_token
        request.session['refresh_token'] = refresh_token

        # If the user can deactivate inactivity, do so.
        user_is_service = user.get('is_service', False)
        inactivity = user.get('inactivity', True)
        if not user_is_service and not inactivity and hasattr(request.session, 'adjust_timeout_for_session'):
            request.session.adjust_timeout_for_session(USER_SESSION_DURATION)

        return []

    # noinspection PyMethodMayBeStatic
    def forget(self, request):
        """Forget the current user.

        Args:
            request (pyramid.request.Request).
        """
        if 'user' in request.session:
            del request.session['user']
        if 'access_token' in request.session:
            del request.session['access_token']
        if 'refresh_token' in request.session:
            del request.session['refresh_token']
        return []


def rights_without_tenants(principals):
    """Transform list of rights in the form (tenant_id, right_name) to list of right_name.

    This is usefull to know if a user has access to an action which does not require tenanting (e.g. list users).
    This function does not remove special Pyramid rights (Authenticated, Everyone).

    Args:
        principals (list of tuples): rights, under the form (tenant_id, right_name).

    Returns:
        set: list of rights names.
    """
    rights = set()

    for principal in principals:
        if isinstance(principal, Right):
            rights.add(principal.name)
        else:
            rights.add(principal)

    return rights


@implementer(IAuthorizationPolicy)
class TenantedAuthorizationPolicy(object):
    """Tenanted Pyramid ACLAuthorizationPolicy. This allows us to define rights tenant by tenant, walling clients
    from one another.
    """

    # noinspection PyMethodMayBeStatic
    def permits(self, context, principals, permission):
        """Copy/Paste of pyramid.authorization.ACLAuthorizationPolicy except for the 2 lines commented.
        Refer to Pyramid for the full doc.

        Basically instead of using the classic Pyramid model of ACL (Allow, principal, permission), we use
        (Allow, tenant, principal, permission). With one catch: some actions do not require tenant (e.g. list objects)
        and the 'admin' principal renders tenanting useless.
        """
        acl = '<No ACL found on any object in resource lineage>'
        principals_without_tenants = rights_without_tenants(principals)

        for location in lineage(context):
            try:
                acl = location.__acl__
            except AttributeError:
                continue

            if acl and callable(acl):
                acl = acl()

            for ace in acl:
                # 4 variables instead of 3.
                ace_action, ace_tenant, ace_principal, ace_permissions = ace
                # If the ACL is not tenanted, only check the right name, otherwise compare the rights with the
                # (tenant_id, right_name) format.
                if (
                    (ace_tenant is None and ace_principal in principals_without_tenants)
                    or Right(name=ace_principal, tenant=ace_tenant) in principals
                ):
                    if not is_nonstr_iter(ace_permissions):
                        ace_permissions = [ace_permissions]
                    if permission in ace_permissions:
                        if ace_action == Allow:
                            return ACLAllowed(ace, acl, permission, principals, location)
                        else:
                            return ACLDenied(ace, acl, permission, principals, location)

        # Default deny (if no ACL in lineage at all, or if none of the principals were mentioned in any ACE we found).
        return ACLDenied('<default deny>', acl, permission, principals, context)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def principals_allowed_by_permission(self, context, permission):
        """Return a set of principal identifiers allowed by the ``permission`` in ``context``.
        Not needed in our case.
        """
        return NotImplementedError()
