"""Celery configuration: set up the Celery app + Celery beat on both sides: 'producer' and 'worker'."""
import json

import click
import sentry_sdk
from celery import Celery, schedules, signals
from pyramid.settings import aslist
from sentry_sdk.integrations.celery import CeleryIntegration
from sentry_sdk.integrations.pyramid import PyramidIntegration
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

from parsys_utilities.config import TenantConfigurator


def configure_celery_app(config_file, sentry=False):
    """Configure the Celery app, both 'producer' and 'worker' sides.

    Args:
        config_file (str): location of the ini config file.
        sentry (bool): should we configure Sentry? Producer side, it is already configured.
    """
    app.conf.tenant_config = TenantConfigurator(config_file)
    config = app.conf.tenant_config.settings

    # Set up the Celery broker (producer/worker communication).
    app.conf.broker_url = config['celery']['broker_url']

    # Add specific apps tasks.
    imports = config['celery'].get('imports')
    if imports:
        app.conf.imports += aslist(imports)

    # Set up Sentry.
    sentry_dsn = config['app:main'].get('sentry.dsn')
    if sentry and sentry_dsn:
        integrations = [CeleryIntegration(), PyramidIntegration(), RedisIntegration(), SqlalchemyIntegration()]
        sentry_sdk.init(sentry_dsn, integrations=integrations, attach_stacktrace=True)

    configure_celerybeat(config)


def configure_celerybeat(config):
    """Add schedule configuration to Celery app from ini file.

    format:
    [celerybeat_<task_name>]
    schedule = <crontab>
    task = <task_callback>
    kwargs = <keyword arguments>  # kwargs are optional

    example:
    [celerybeat_notify-every-morning]
    schedule = {'hour': 7, 'minute': 30}
    task = parsys_cloud.celery_tasks.checkup_validity_reminder
    kwargs = {'validity_months': 3}

    Args:
        config (configparser.ConfigParser).
    """
    beat_config = {}
    celerybeat_sections = [section for section in config.sections() if section.startswith('celerybeat:')]
    if not celerybeat_sections:
        return

    for section in celerybeat_sections:
        task_parameters = {
            'schedule': schedules.crontab(**json.loads(config[section]['schedule'])),
            'task': config[section]['task'],
        }

        # Kwargs are optional.
        if 'kwargs' in config[section]:
            task_parameters['kwargs'] = json.loads(config[section]['kwargs'])

        task_name = section[len('celerybeat:'):]
        beat_config[task_name] = task_parameters

    app.conf.beat_schedule = beat_config


# noinspection PyUnusedLocal
@signals.user_preload_options.connect
def on_preload_parsed(options, **kwargs):
    """Configure the Celery app on worker/beat startup. We require the 'ini' option.

    Args:
        options (dict).
    """
    configure_celery_app(options['ini'], sentry=True)


# Create the Celery app on both 'producer' and 'worker' sides.
app = Celery()
# Add the --ini option to the worker.
ini_type = click.Path(exists=True, file_okay=True, dir_okay=False, readable=True)
app.user_options['preload'].add(
    click.Option('--ini', required=True, type=ini_type, help='App ini configuration file.')
)

# Import parsys_utilities tasks by default.
app.conf.imports = ['parsys_utilities.celery_tasks']
