import nexmo
from requests import post, RequestException
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from parsys_utilities.celery_app import app


def get_session_factory():
    """Create an SQLAlchemy session factory.

    Returns:
        sqlalchemy.orm.sessionmaker.
    """
    sqlalchemy_url = app.conf.tenant_config.settings['app:main']['sqlalchemy.url']
    engine = create_engine(sqlalchemy_url)
    return sessionmaker(bind=engine)


@app.task(bind=True)
def post_json(self, server_url, json, username=None, password=None):
    try:
        auth = (username, password) if username and password else None
        headers = {'accept': 'application/json'}
        r = post(server_url, json=json, auth=auth, headers=headers)
        return r.status_code

    except RequestException as exc:
        self.retry(countdown=30, exc=exc, max_retries=5)


@app.task(bind=True)
def post_mailgun_email(self, server_url, auth_key, **kwargs):
    try:
        headers = {'accept': 'application/json'}
        r = post(server_url, auth=('api', auth_key), headers=headers, **kwargs)
        return r.status_code

    except RequestException as exc:
        self.retry(countdown=30, exc=exc, max_retries=5)


@app.task(bind=True)
def post_nexmo_sms(self, api_key, api_secret, sender, recipient, text):
    client = nexmo.Client(key=api_key, secret=api_secret)
    response = client.send_message({'from': sender, 'to': recipient, 'text': text})

    try:
        response = response['messages'][0]

        if response['status'] == '0':
            return response['message-id']
        else:
            raise RequestException(response['error-text'])

    except (KeyError, RequestException) as exc:
        self.retry(countdown=30, exc=exc, max_retries=5)
