from configparser import ConfigParser
from pathlib import Path
from urllib.parse import urlparse

from pyramid.settings import aslist_cronly


def aslist(value, flatten=True):
    """ Return a list of strings, separating the input based on newlines
    and, if flatten=True (the default), also split on spaces within
    each line."""
    if value is None or value == '':
        return []

    values = aslist_cronly(value)
    if not flatten:
        return values
    result = []
    for value in values:
        subvalues = value.split()
        result.extend(subvalues)
    return result


class TenantConfigurator(object):
    def __init__(self, config_file, defaults=None):
        # Parse the ini config file.
        self.settings = ConfigParser()
        self.settings.read(config_file)
        # Look for specific config files.
        config_folder = Path(config_file).parent
        for specific_file in config_folder.glob('specific/*.ini'):
            self.settings.read(specific_file)

        self.defaults = defaults or {}

    def get(self, parameter, section_name, main=True, empty_accepted=False):
        """Returns configuration value in accordance with the wanted key (domain or tenant).

        Args:
            parameter (str).
            section_name (str).
            main (bool): if no direct section is found, should we return the main/default?
            empty_accepted (bool): is an empty value considered valid?

        Returns:
            str: configuration value.
        """
        sections = self.settings.sections()

        # Direct match: 'tenant:kv56Zf4n'.
        if section_name in sections:
            setting = self.settings[section_name].get(parameter)
            if (empty_accepted and setting == '') or setting:
                return setting

        if not main:
            return None

        # No match: use global configuration.
        setting = self.settings['app:main'].get(parameter)
        if (empty_accepted and setting == '') or setting:
            return setting

        # Parameter isn't present in ini file, use default.
        return self.defaults.get(parameter)

    def get_for_all_tenants(self, parameter):
        """Get the configuration for all tenants. Useful for admins.

        Args:
            parameter (str): wanted key.

        Returns:
            dict (str:str): configuration, each key being a tenant + 'default' .
        """
        output = {
            'default': self.settings['app:main'].get(parameter, ''),
        }

        for section in self.settings.sections():
            if not section.startswith('tenant:'):
                continue

            section_ids = section[section.find(':') + 1:].split(',')
            for section_id in section_ids:
                output[section_id] = self.settings[section].get(parameter, '')

        return output

    def get_for_domain(self, parameter, url, empty_accepted=False):
        """Get the configuration for one parameter for a given domain.

        Args:
            parameter (str): wanted key.
            url (str): request url.
            empty_accepted (bool): is an empty value acceptable? Useful if the default is a value and we want to
                deactivate this for a given domain (e.g. modules/specific).

        Returns:
            str: configuration for the given domain.
        """
        domain = urlparse(url).netloc
        return self.get(parameter, f'domain:{domain}', empty_accepted=empty_accepted)

    def get_for_tenant(self, parameter, tenant, main=True, empty_accepted=False):
        """Get the configuration for one parameter for a given tenant.

        Args:
            parameter (str): wanted key.
            tenant (str): wanted tenant.
            main (bool): if no direct section is found, should we return the main/default?
            empty_accepted (bool): is an empty value acceptable? Useful if the default is a value and we want to
                deactivate this for a given tenant (e.g. modules/specific).

        Returns:
            str: configuration for the given tenant.
        """
        return self.get(parameter, f'tenant:{tenant}', main=main, empty_accepted=empty_accepted)
