from collections import OrderedDict
from datetime import datetime

import arrow
import babel.dates
import l18n
from pytz import common_timezones
from unidecode import unidecode

# Simple regions translations for timezones.
TIMEZONES_REGIONS = {
    'fr': {
        'Africa': 'Afrique',
        'America': 'Amérique',
        'Antarctica': 'Antarctique',
        'Arctic': 'Arctique',
        'Asia': 'Asie',
        'Australia': 'Australie',
        'Canada': 'Canada',
        'Europe': 'Europe',
        'Indian': 'Indien',
        'Pacific': 'Pacifique',
        'US': 'États-Unis',
    }
}


def format_date(date, locale, date_format='medium'):
    if isinstance(date, str):
        try:
            date = datetime.strptime(date, '%Y-%m-%d')
        except ValueError:
            return date

    return babel.dates.format_date(date=date, format=date_format, locale=locale)


def format_datetime(date_time, locale, timezone, date_format='medium', naive=True):
    if isinstance(date_time, str):
        try:
            date_time = arrow.get(date_time)
        except arrow.parser.ParserError:
            return date_time

    date_time = babel.dates.format_datetime(datetime=date_time, format=date_format, tzinfo=timezone, locale=locale)
    if date_format in ['short', 'medium'] and not naive:
        offset = babel.dates.format_datetime(format='ZZZ', tzinfo=timezone, locale=locale)
        date_time = f'{date_time} {offset}'
    return date_time


def localize_timezones(locale):
    """Create a dict of timezones with the format iso name: translated label.

    Args:
        locale (str): user's locale.

    Returns:
        OrderedDict: translated timezones.
    """
    # Set locale.
    l18n.set_language(locale)

    translated_timezones = {}
    for timezone in common_timezones:
        # Get timezone translation using l18n.
        translated_timezones[timezone] = str(l18n.tz_fullnames.get(timezone, ''))

        # There are rare cases where l18n doesn't have the translation. In these cases, at least translate the region
        # so that the timezone isn't too weirdly placed.
        if not translated_timezones[timezone] and TIMEZONES_REGIONS.get(locale):
            translated_timezone = timezone
            for region in TIMEZONES_REGIONS[locale].items():
                translated_timezone = translated_timezone.replace(*region)

            translated_timezones[timezone] = translated_timezone

        # If we didn't create a regions translation for the user's locale, just use the untranslated timezone.
        if not translated_timezones[timezone]:
            translated_timezones[timezone] = timezone

    # Sort timezones according to translated label.
    return OrderedDict(sorted(translated_timezones.items(), key=lambda x: unidecode(x[1])))


def parse_date(date_input):
    """Get python date from form format: YYYY-MM-DD.

    Args:
        date_input (str): date in YYYY-MM-DD.

    Returns:
        datetime.date if date is valid, ValueError otherwise.
    """
    if date_input and isinstance(date_input, str):
        return datetime.strptime(date_input, '%Y-%m-%d').date()
