"""Logging utilities: add user ip/session id/user id to every log."""
from logging import getLogger, LoggerAdapter


class ContextAdapter(LoggerAdapter):
    """LoggerAdapter that adds user ip/session id/user id to every log."""

    def process(self, msg, kwargs):
        """Process the logging message.

        Args:
            msg (str or list).
            kwargs (dict): only expected is "user_id", the rest will be passed along.
        """
        # If we want to log more than one info (a dict), transform it into a string.
        if isinstance(msg, list):
            msg = ','.join([str(string) for string in msg])

        # By allowing to override the user_id at log time, we can retrieve this info that sometime isn't available at
        # the beginning of the request but is at its end (ie: in all login situations).
        if kwargs.get('user_id'):
            self.extra['user_id'] = kwargs['user_id']
            kwargs.pop('user_id')

        return f'{self.extra["ip"]},{self.extra["session_id"]},{self.extra["user_id"]},{msg}', kwargs


def logger(request, name=__package__):
    """This function is used by the apps to create our ContextAdapter according to the request and store it for future
    use.
    """
    app_logger = getLogger(name)

    # User id name is different between RTA and OAuth clients.
    user_id = None
    if request.user and hasattr(request.user, 'user_id'):
        user_id = request.user.user_id
    elif request.user:
        user_id = request.user.id

    # Try to get the user's ip:
    #   - first from Nginx (X-Real-IP).
    #   - then using request.client_addr. Note you did it this way because this method uses request headers, which
    # means it's user-declared and is super easy to fake.
    client_addr = request.headers.get('x-real-ip', f'CA+{request.client_addr}')

    extra = {'ip': client_addr, 'session_id': request.session.ensure_id(), 'user_id': user_id}
    return ContextAdapter(app_logger, extra)


def utilities_logger(request):
    """Little function to make sure that logging happening in parsys_utilities doesn't blow-up if the app using it
    doesn't respect our contract. Normally, apps will have a "logger_actions". In the weird case that they don't,
    create a logger.
    """
    if hasattr(request, 'logger_actions'):
        return request.logger_actions
    else:
        return logger(request)
