from inflection import underscore
from sqlalchemy import case, Column, DateTime, func, Integer, Unicode as String
from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import Mutable, MutableList
from sqlalchemy.schema import MetaData
from sqlalchemy.types import TypeDecorator


class _ModelBase(object):
    id = Column(Integer, primary_key=True)


class _DeclarativeMeta(DeclarativeMeta):
    def __new__(mcs, name, bases, d):
        d['__tablename__'] = d.get('__tablename__') or d.get('__table__') or underscore(name)
        return DeclarativeMeta.__new__(mcs, name, bases, d)


# Recommended naming convention used by Alembic, as various different database
# providers will autogenerate vastly different names making migrations more
# difficult. See: http://alembic.readthedocs.org/en/latest/naming.html
NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

metadata = MetaData(naming_convention=NAMING_CONVENTION)

Model = declarative_base(cls=_ModelBase, name='ObjectModel', metaclass=_DeclarativeMeta, metadata=metadata)


class AliasMixin(object):
    first_name = Column(String)
    last_name = Column(String)

    @hybrid_property
    def full_name(self):
        if self.last_name and self.first_name:
            return f'{self.last_name} {self.first_name}'
        elif self.last_name:
            return self.last_name
        elif self.first_name:
            return self.first_name

    # noinspection PyMethodParameters,PyUnresolvedReferences
    @full_name.expression
    def full_name(cls):
        return case([
            (cls.last_name.isnot(None) & cls.first_name.isnot(None), cls.last_name + ' ' + cls.first_name),
            (cls.last_name.isnot(None), cls.last_name),
            (cls.first_name.isnot(None), cls.first_name)
        ],
            else_=None,
        )


class CreationDateTimeMixin(object):
    created_at = Column(DateTime, default=func.now(), nullable=False)


class UpdateDateTimeMixin(object):
    updated_at = Column(DateTime, default=func.now(), onupdate=func.now(), nullable=False)


# noinspection PyArgumentList
class MutableListAcceptStr(MutableList):
    @classmethod
    def coerce(cls, index, value):
        if isinstance(value, cls):
            return value

        if isinstance(value, list):
            return cls(value)

        elif isinstance(value, str):
            return cls([value])

        return Mutable.coerce(index, value)


# noinspection PyAbstractClass
class ScalarList(TypeDecorator):
    """Flatten lists to store them as strings, get list back when accessing the attribute.
    Refuse strings using our separator: '###'.
    """

    impl = String

    def process_bind_param(self, value, dialect):
        # We can't do 'if not value' as we should be able to accept '0';
        if value is None or value == '' or value == []:
            return

        # This is mandatory for SQL searches.
        if not isinstance(value, list):
            value = [value]

        # Remove empty strings.
        value = [item for item in value if item != '']

        # Refuse str using our separator.
        if any('###' in str(item) for item in value):
            raise ValueError()

        # Add separators at the beginning and the end to make sure that 'contains' doesn't match on partial values.
        return f'###{"###".join(str(item) for item in value)}###'

    def process_result_value(self, value, dialect):
        if not value:
            return []

        # Remove first and last element which are empty strings created by our extra separators.
        return value.split('###')[1:-1]


ScalarList = MutableListAcceptStr.as_mutable(ScalarList)
