"""Simple notifier to send notifications from the OAuth2 apps to RTA, which will take care of dispatching to the right
user.
"""
import gettext
from urllib.parse import urljoin

from jinja2 import Environment, FileSystemLoader
from pyramid.i18n import make_localizer
from sentry_sdk import capture_exception

from parsys_utilities import AVAILABLE_LOCALES, dates
from parsys_utilities.celery_tasks import post_json


def validate_json(json, external):
    """Validate notifications json:
        - must have message and tenant
        - must have either rights or user_ids

    Args:
          json(dict).
          external(bool): if the notification is external, no need to validate rights/user ids.
    """
    if not json.get('message') or not json.get('tenant'):
        raise ValueError()

    if not external and not json.get('rights') and not json.get('user_ids'):
        raise ValueError()


class Notifier(object):
    def __init__(self, request, send_notifications=True):
        self.request = request
        self.send_notifications = send_notifications

    def notify(self, json, external=False):
        """Send notifications to users: send the request to RTA, who will take care of dispatching to the right users.

        Notifications are sent asynchronously via Celery to be processed by RTA.
        To RTA, we send: the users' tenant + either the users' ids or specific rights.

        Args:
            json (dict): email content.
            external (boolean): whether the e-mail should be sent with RTA external notification API.

        json structure:
            message (dict): {'email': {<locale_name>: {'subject': str, 'text': str, 'html': str}}}.
            tenant (str).
            user_ids or rights (list(str)).
            level (str).
        """
        try:
            validate_json(json, external)
        except ValueError as error:
            capture_exception(error)
            return

        if not self.send_notifications:
            return

        client_id = self.request.registry.settings['rta.client_id']
        secret = self.request.registry.settings['rta.secret']

        rta_domain = self.request.registry.tenant_config.get_for_domain('rta.server_url', self.request.host_url)
        notifications_url = urljoin(rta_domain, '/api/notifications/external/' if external else '/api/notifications/')

        post_json.apply_async(args=[notifications_url, json, client_id, secret])


def notify_offline(tenant_config, json, external=False):
    """Send notifications from outside the WGSI app.

    This function is a modification of Notifier.notify().

    Args:
        tenant_config (parsys_utilities.config.TenantConfigurator).
        json (dict): email content.
        external (boolean): whether the e-mail should be sent with RTA external notification API.

    json structure:
        message (dict): {'email': {<locale_name>: {'subject': str, 'text': str, 'html': str}}}.
        tenant (str).
        user_ids or rights (list(str)).
        level (str).
    """
    validate_json(json, external)

    client_id = tenant_config.settings['app:main']['rta.client_id']
    secret = tenant_config.settings['app:main']['rta.secret']

    rta_server_url = tenant_config.get_for_tenant('rta.server_url', json['tenant'])
    notifications_url = urljoin(rta_server_url, '/api/notifications/external/' if external else '/api/notifications/')

    post_json.apply_async(args=[notifications_url, json, client_id, secret])


def emails_renderer_offline(templates_path, translations_path, subject, text_file, html_file, template_data):
    """Generates emails in all the languages available in this application.

    RTA will choose the language according to the user parameters.

    This function is a modification of emails_renderer().
    It allows us to send notifications from outside the WGSI app, in our case, from a Celery worker.

    Args:
        templates_path (str).
        translations_path (str).
        subject (str): email subject.
        text_file (str): path to message template as plain text.
        html_file (str): path to message template as html.
        template_data (dict).

    Returns:
        dict: messages {<locale_name>: {'subject': str, 'text': str, 'html': str}}.
    """
    # jinja2 configuration
    env = Environment(
        extensions=['jinja2.ext.i18n', 'jinja2.ext.autoescape', 'jinja2.ext.with_'],
        loader=FileSystemLoader([templates_path]),
    )
    env.filters['format_date'] = dates.format_date  # add custom filter to format date according to locale

    text_template = env.get_template(text_file)
    html_template = env.get_template(html_file)

    emails = {}

    for locale in AVAILABLE_LOCALES:
        # translate subject
        pyramid_localizer = make_localizer(current_locale_name=locale, translation_directories=[translations_path])
        rendered_subject = pyramid_localizer.translate(subject)

        # translate template
        template_data['locale'] = locale
        translations = gettext.translation(domain='messages', localedir=translations_path, languages=[locale])
        env.install_gettext_translations(translations, newstyle=True)
        rendered_text = text_template.render(template_data)
        rendered_html = html_template.render(template_data)

        emails[locale] = {
            'subject': rendered_subject,
            'text': rendered_text,
            'html': rendered_html
        }

    return emails
