import time
from datetime import datetime, timedelta
from json import JSONDecodeError
from urllib.parse import parse_qs, urlencode, urljoin, urlparse

import jwt
import requests
from oauthlib.oauth2 import OAuth2Error, WebApplicationClient
from pyramid.csrf import get_csrf_token
from pyramid.httpexceptions import HTTPBadRequest, HTTPFound, HTTPUnauthorized, HTTPServiceUnavailable
from pyramid.renderers import render_to_response
from pyramid.security import Allow, Authenticated, forget, remember
from pyramid.view import forbidden_view_config, view_config
# noinspection PyProtectedMember
from requests.auth import _basic_auth_str
from sentry_sdk import capture_message

from parsys_utilities.logs import utilities_logger


def validate_referer(referer, host_url):
    """Make sure that the referer URL is from the same domain than the server, so that users can't be maliciously
    redirected after login.

    Args:
        referer (str): request referer.
        host_url (str): request host URL.

    Returns:
        validated URL or '/' (str).
    """
    parsed_referer = urlparse(referer)
    parsed_host_url = urlparse(host_url)
    if (parsed_referer.scheme, parsed_referer.netloc) == (parsed_host_url.scheme, parsed_host_url.netloc):
        return referer
    else:
        return '/'


def validate_response(response):
    if response.status_code == 200:
        try:
            return response.json()
        except JSONDecodeError:
            capture_message('RTA error.')
            raise HTTPBadRequest()
    elif response.status_code == 400:
        raise HTTPBadRequest()
    elif response.status_code in {401, 403}:
        raise HTTPUnauthorized()
    else:
        capture_message('RTA error.')
        raise HTTPServiceUnavailable()


def request_user_info(request, access_token, auth):
    rta_domain = request.registry.tenant_config.get_for_domain('rta.server_url', request.host_url)
    rta_userinfo_url = urljoin(rta_domain, '/api/oauth/userinfo/')
    headers = {
        'accept': 'application/json',
        'authorization': f'Bearer {access_token}',
        'X-OpenID-Origin': _basic_auth_str(*auth),
    }

    try:
        response = requests.get(rta_userinfo_url, headers=headers)
    except requests.RequestException:
        raise HTTPServiceUnavailable()

    return validate_response(response)


class OpenID(object):
    def __init__(self, request):
        self.request = request
        self.openid_client_id = self.request.registry.tenant_config.get_for_domain(
            'rta.client_id', self.request.host_url
        )
        self.openid_client_secret = self.request.registry.tenant_config.get_for_domain(
            'rta.secret', self.request.host_url
        )
        self.openid_client_auth = (self.openid_client_id, self.openid_client_secret)
        self.openid_client = WebApplicationClient(self.openid_client_id)

    def verify_authorization_response(self, url, session_csrf_token):
        try:
            # Retrieve information from Oauth authorization success.
            authorization_response = self.openid_client.parse_request_uri_response(url)
            authorization_code = authorization_response['code']
            state_encoded = authorization_response['state']
            state = parse_qs(state_encoded)

            state_csrf_token = state['csrf_token'][0]
            user_target_path = state['continue'][0]
            user_target_path = urlparse(user_target_path).path
        except (IndexError, KeyError, TypeError):
            raise HTTPBadRequest()

        if state_csrf_token != session_csrf_token:
            raise HTTPBadRequest()

        return authorization_code, user_target_path

    def verify_json_web_token(self, token, issuer):
        now = datetime.utcnow()

        try:
            signed_id_token = token['id_token']
            access_token_expires_at = now + timedelta(seconds=int(token['access_token_expires_in']))
            access_token = {'token': token['access_token'], 'expires_at': access_token_expires_at}
            refresh_token_expires_at = now + timedelta(seconds=int(token['refresh_token_expires_in']))
            refresh_token = {'token': token['refresh_token'], 'expires_at': refresh_token_expires_at}
        except (KeyError, TypeError):
            raise HTTPBadRequest()

        try:
            # Even if we don't persist it, make sure that the id_token is valid.
            key = self.openid_client_secret
            jwt.decode(signed_id_token, key, algorithms=['HS256'], audience=self.openid_client_id, issuer=issuer)
        except (jwt.ExpiredSignatureError, jwt.InvalidAudienceError, jwt.InvalidIssuerError, OAuth2Error):
            raise HTTPBadRequest()
        else:
            return access_token, refresh_token

    def request_access_token_with_authorization_code(self, authorization_code, client_redirect_endpoint):
        # Using authorization code, request token.
        rta_domain = self.request.registry.tenant_config.get_for_domain('rta.server_url', self.request.host_url)
        rta_token_url = urljoin(rta_domain, '/api/oauth/token/')

        data = self.openid_client.prepare_request_body(code=authorization_code, redirect_uri=client_redirect_endpoint)
        headers = {'accept': 'application/json'}
        try:
            response = requests.post(rta_token_url, data=data, auth=self.openid_client_auth, headers=headers, timeout=3)
        except requests.RequestException:
            raise HTTPServiceUnavailable()

        if response.status_code == 400:
            raise HTTPBadRequest()
        elif response.status_code != 200:
            raise HTTPServiceUnavailable()

        return response.json()

    def request_access_token_with_refresh_token(self, refresh_token):
        rta_domain = self.request.registry.tenant_config.get_for_domain('rta.server_url', self.request.host_url)
        rta_token_url = urljoin(rta_domain, '/api/oauth/token/')

        data = self.openid_client.prepare_refresh_body(refresh_token=refresh_token)
        headers = {'accept': 'application/json'}
        try:
            response = requests.post(rta_token_url, data=data, auth=self.openid_client_auth, headers=headers, timeout=3)
        except requests.RequestException:
            raise HTTPServiceUnavailable()

        json_web_token = validate_response(response)
        return self.verify_json_web_token(json_web_token, rta_domain)

    @forbidden_view_config()
    def manage_forbidden(self):
        if self.request.user:
            self.request.response.status_int = 403
            return render_to_response('errors/403.html', {}, request=self.request, response=self.request.response)

        is_medcapture = self.request.user_agent and self.request.user_agent.startswith('ParsysMedicalStation')
        is_medcapture_api = is_medcapture and self.request.path.startswith('/api/')
        # is_xhr: pjax + datatables.
        if self.request.is_xhr or is_medcapture_api:
            return HTTPUnauthorized()

        if self.request.method in ['DELETE', 'POST', 'PUT']:
            continue_url = validate_referer(self.request.referer, self.request.host_url)
        else:
            continue_url = self.request.path
        state = {'continue': continue_url, 'csrf_token': get_csrf_token(self.request)}
        parameters = {
            'now': int(time.time()),
            'redirect_uri': self.request.route_url('oauth-token'),
            'scope': ['openid'],
            'state': urlencode(state),
        }

        rta_domain = self.request.registry.tenant_config.get_for_domain('rta.server_url', self.request.host_url)
        rta_authorization_url = urljoin(rta_domain, '/api/oauth/authorize/')

        rta_authorization_request = self.openid_client.prepare_request_uri(rta_authorization_url, **parameters)
        return HTTPFound(location=rta_authorization_request)

    @view_config(route_name='oauth-token', request_method='GET')
    def request_token(self):
        csrf_token = get_csrf_token(self.request)
        authorization_code, user_target_path = self.verify_authorization_response(self.request.url, csrf_token)
        request_token_endpoint = self.request.route_url('oauth-token')
        json_web_token = self.request_access_token_with_authorization_code(authorization_code, request_token_endpoint)

        issuer = self.request.registry.tenant_config.get_for_domain('rta.server_url', self.request.host_url)
        access_token, refresh_token = self.verify_json_web_token(json_web_token, issuer)
        user_info = request_user_info(self.request, access_token['token'], self.openid_client_auth)

        remember(self.request, user_info['id'], user=user_info, access_token=access_token, refresh_token=refresh_token)

        utilities_logger(self.request).info('log in {}'.format(user_info['id']))
        return HTTPFound(location=user_target_path)


class Users(object):
    __acl__ = [
        (Allow, None, Authenticated, 'authenticated'),
    ]

    def __init__(self, request):
        self.request = request
        self.rta_server_url = self.request.registry.tenant_config.get_for_domain(
            'rta.server_url', self.request.host_url
        )

    @view_config(route_name='users-logout', request_method='GET')
    def logout_get(self):
        utilities_logger(self.request).info('log out')

        forget(self.request)

        host_url = self.request.host_url
        rta_logout_url = urljoin(self.rta_server_url, f'/users/logout/?continue={host_url}')
        return HTTPFound(location=rta_logout_url)

    @view_config(route_name='users-settings', request_method='GET', permission='authenticated',
                 renderer='pages/settings.html')
    def settings_get(self):
        template_vars = {'rta_settings_url': urljoin(self.rta_server_url, '/users/settings-iframe/')}
        response = render_to_response(
            'pages/settings.html', template_vars, request=self.request, response=self.request.response
        )
        # Pop session user AFTER rendering the page so that the data isn't fetched from RTA again.
        self.request.session.pop('user')
        return response


def includeme(config):
    config.add_route(pattern='oauth/token/', name='oauth-token')
    config.add_route(pattern='users/logout/', name='users-logout')
    config.add_route(pattern='users/settings/', name='users-settings', factory=Users)
