from pyramid.settings import asbool
from sqlalchemy import asc, Boolean, cast, desc, func, literal, or_, select, Unicode as String, union_all
from sqlalchemy.exc import InvalidRequestError


def sql_filter(q, searched_object, specific_attributes, search):
    """Filter a query.

    Args:
        q (sqlalchemy.orm.query.Query).
        searched_object (sqlalchemy.ext.DeclarativeMeta).
        specific_attributes (dict).
        search (list).

    Returns:
        sqlalchemy.orm.query.Query.
    """
    search = search or []

    for search_element in search:
        if search_element[0] in specific_attributes:
            attribute = specific_attributes[search_element[0]]

        elif hasattr(searched_object, search_element[0]):
            attribute = getattr(searched_object, search_element[0])

        else:
            raise KeyError(f'Unknown attribute: {search_element[0]}.')

        search_variable = search_element[2]
        try:
            # Transform pseudo Booleans (like 1 or 'True') into real Booleans.
            if search_variable in ['false', 'true', 'False', 'True'] or isinstance(attribute.type, Boolean):
                search_variable = asbool(search_element[2])
        except (AttributeError, InvalidRequestError):
            pass

        if search_element[1] == '==':
            q = q.filter(attribute == search_variable)
        elif search_element[1] == '!=':
            q = q.filter(attribute != search_variable)
        elif search_element[1] == 'contains' and search_variable is not None:
            q = q.filter(attribute.contains(search_variable))
        elif search_element[1] == 'in' and search_variable is not None:
            q = q.filter(attribute.in_(search_variable))

    return q


def sql_search(db_session, searched_object, full_text_search_attributes, joined_tables=None, tenanting=None,
               specific_attributes=None, search_parameters=None, options=None):
    """Perform a standard SQL query on our DB.

    To cater to all special needs, the infos of the queried object can be specified (full text search, special sorts,
    etc.).
    Search parameters:
        full_text_search (str).
        search (list of tuples): Tuples are (object attribute, search value).
        sort (list of str): Each element should be an attribute of the object or defined in the specific sort
        attributes.
        Sorting is done in the order of the elements in the list. Precede the element by '-' to do a descending sort.

    Args:
        db_session (sqlalchemy.session.Session): Database session.
        searched_object (sqlalchemy.ext.DeclarativeMeta): Objects type to search (sqlalchemy.schema.Table).
        full_text_search_attributes (list of sqlalchemy.Columns): Object's attributes to search on during full text
        search.
        joined_tables (List(sqlalchemy.ext.DeclarativeMeta)): Tables to join on (for full text search/tenanting/search).
        tenanting (function sqlalchemy.query.query:sqlalchemy.query.query): Function which must be applied to return
        tenanted list.
        specific_attributes (dict): SQLAlchemy column to search/sort on when the search term doesn't match the
        corresponding object's attribute.
        search_parameters (dict): Limit/offset/search term/sort/full text search.
        options (list).

    Returns:
        dict:
            items (int): Searched objects.
            recordsTotal (int): Total number of objects after tenanting.
            recordsFiltered (int): Number of filtered objects (because of search/full text search).
    """
    joined_tables = joined_tables or []
    search_parameters = search_parameters or {}
    specific_attributes = specific_attributes or []
    output = {}

    q = db_session.query(searched_object)
    for table in joined_tables:
        q = q.outerjoin(table)

    # Apply tenanting first. Users should only know the number of objects for their tenant.
    if tenanting:
        q = tenanting(q)

    output['recordsTotal'] = q.count()

    sort = search_parameters.get('sort') or []
    q = sql_sort(q, searched_object, specific_attributes, sort)
    search = search_parameters.get('search') or []
    q = sql_filter(q, searched_object, specific_attributes, search)

    # Apply full text search.
    if search_parameters.get('full_text_search'):
        for word in search_parameters['full_text_search'].split(' '):
            search_filter = [attribute.ilike(f'%{word}%') for attribute in full_text_search_attributes]
            q = q.filter(or_(*search_filter))

    # Number of records after filters.
    output['recordsFiltered'] = q.count()

    if search_parameters.get('limit'):
        q = q.limit(search_parameters['limit'])

    if search_parameters.get('offset'):
        q = q.offset(search_parameters['offset'])

    if options:
        q = q.options(*options)

    output['items'] = q.all()
    return output


def sql_sort(q, searched_object, specific_attributes, sort):
    """Sort a query.

    Args:
        q (sqlalchemy.orm.query.Query).
        searched_object (sqlalchemy.ext.DeclarativeMeta).
        specific_attributes (dict).
        sort (list).

    Returns:
        sqlalchemy.orm.query.Query.
    """
    for sort_element in sort:
        if sort_element.startswith('-'):
            sort_function = desc
            sort_element = sort_element.lstrip('-')
        else:
            sort_function = asc

        special_sort = f'/sort/{sort_element}'
        if special_sort in specific_attributes:
            q = q.order_by(sort_function(specific_attributes[special_sort]))

        elif sort_element in specific_attributes:
            q = q.order_by(sort_function(specific_attributes[sort_element]))

        elif hasattr(searched_object, sort_element):
            attribute = getattr(searched_object, sort_element)
            try:
                # When sorting strings, always try to do a case-insensitive sort.
                if isinstance(attribute.type, String):
                    q = q.order_by(sort_function(func.lower(attribute)))
                else:
                    q = q.order_by(sort_function(attribute))
            # Hybrid properties will fail here, so do a simple sort.
            except (AttributeError, InvalidRequestError):
                q = q.order_by(sort_function(attribute))

        # If we didn't find the attribute to sort on, fail.
        else:
            raise KeyError(f'Unknown attribute: {sort_element}.')

    # Always sort by ID last. Otherwise when we sort by non unique columns, we cannot guarantee the pagination will be
    # correct.
    q = q.order_by(searched_object.id)

    return q


def table_from_dict(table_name, data):
    """Create a Common Table Expression (CTE) based on a Python dict so that it can be joined with tables when we
    are SQL sorting/filtering. This is useful when we have data in Python that matches with tables data (translations,
    tenants names <=> tenants ids, etc.).
    """
    stmts = [
        select([
            cast(literal(item[key]), String).label(key)
            for key in item.keys()
        ])

        for item in data
    ]

    subquery = union_all(*stmts)
    return subquery.cte(name=table_name)
