"""Provide a docker healthcheck and a status endpoint to Python apps, testing persistence, redis and celery."""

import argparse
import collections
from datetime import datetime
from random import randint
from traceback import format_exc

import celery.bin.base
import pkg_resources
import requests
import requests_unixsocket
from pyramid.settings import asbool
from pyramid.view import exception_view_config, notfound_view_config, view_config
from redis.exceptions import RedisError
from sqlalchemy.exc import SQLAlchemyError

from parsys_utilities.celery_app import app


def healthcheck():
    """Command-line utility. Check if the docker container is healthy.
    In dev, we expose a port, in prod, we use a socket.

    If the get fails, an exception will be raised and the command return will be != 0.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--dev', action='store_true')
    args, extras = parser.parse_known_args()

    if args.dev:
        requests.get('http://localhost:80/health/', timeout=0.5)
    else:
        requests_unixsocket.get('http+unix://%2Fsrv%2Fweb.sock/health/', timeout=0.5)


def get_module_version(package_name):
    """[serverVersion] Get the version of the server module.

    Args:
        package_name (str).

    Returns:
        str: app version.
    """
    try:
        version = pkg_resources.require(package_name)[0].version

    except (pkg_resources.DistributionNotFound, IndexError):
        return None
    else:
        return version


def db_status(db_session):
    """[persistenceService] Check if the database is accessible.

    Args:
        db_session (sqlalchemy.orm.session.Session): sqlalchemy db session.

    Returns:
        bool: is the database working?
    """
    try:
        db_session.execute('SELECT * FROM alembic_version')

    except SQLAlchemyError:
        return False
    else:
        return True


def celery_status():
    """[queueService] Check if celery is running.

    Returns:
        bool: is celery working?
    """
    try:
        celery_inspect = app.control.inspect()

    except celery.bin.base.Error:
        return False
    else:
        # If no worker is running, stats() returns None.
        return celery_inspect.stats() is not None


def redis_status(session):
    """[keyValueService] Validate redis is running.

    Args:
        session (pyramid_redis_sessions.session.RedisSession).

    Returns:
        bool: is redis working?
    """
    test_value = randint(1, 500)

    try:
        session['usage-test-only'] = test_value
        assert session.get('usage-test-only') == test_value
        del session['usage-test-only']

    except (AssertionError, RedisError):
        return False
    else:
        return True


@view_config(route_name='health', request_method='GET', renderer='json')
def health_get(_request):
    """Simple docker healthcheck."""
    return 'healthy'


@view_config(route_name='status', request_method='GET', renderer='json')
def status_get(request):
    """Display app status."""
    status = collections.OrderedDict()
    status['serverVersion'] = get_module_version(request.registry.package_name)
    status['parsysUtilitiesVersion'] = get_module_version('parsys_utilities')
    status['webService'] = True

    # Test if the database is accessible.
    status['persistenceService'] = db_status(request.db_session)

    # Test if redis is accessible.
    status['keyValueService'] = redis_status(request.session)

    # Test celery.
    status['queueService'] = celery_status()

    # If any test is False, return error.
    if any(not value for value in status.values()):
        request.response.status_int = 500

    return status


@notfound_view_config(append_slash=True, renderer='errors/404.html')
def not_found_get(request):
    """Not found page."""
    request.response.status_int = 404
    return {}


@exception_view_config(Exception, renderer='errors/500.html')
def exception(request):
    """Catch exceptions.
    In dev reraise them to be caught by pyramid_debugtoolbar/sentry.
    In production log them then return a 500 page to the user, the exception is automatically caught by sentry.
    """
    # In dev.
    package_name = request.registry.package_name
    debug_exceptions = asbool(request.registry.settings.get(f'{package_name}.dev.debug_exceptions', False))
    if debug_exceptions:
        raise request.exception

    # In production.
    else:
        error_header = f'Time: {datetime.utcnow()}\nUrl: {request.url}\nMethod: {request.method}\n'
        error_text = error_header + format_exc()
        request.logger_technical.error(error_text)

        request.response.status_code = 500
        return {}


def includeme(config):
    config.add_route(pattern='health/', name='health')
    config.add_route(pattern='status/', name='status')
