import os
import shutil
import tempfile
import unittest
from collections import namedtuple
from configparser import ConfigParser
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import partial
from unittest.mock import patch

import alembic.command
import alembic.config
from pyramid.security import Authenticated, Everyone
from pyramid.testing import DummyRequest
from pyramid_session_redis import RedisSessionFactory
from pyramid_session_redis.tests import DummyRedis
from webtest import TestApp

from parsys_utilities import as_list
from parsys_utilities.model import Model

os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'

User = namedtuple('User', ['is_admin'])


@contextmanager
def dummy_context_manager():
    yield None


class DummyAuthenticationPolicy(object):
    def __init__(self, principals=None, **_kw):
        self.userid = 'XXXXXXXX'
        self.principals = principals

    def authenticated_userid(self, _request):
        if self.principals:
            return self.userid

    def unauthenticated_userid(self, request):
        if self.principals:
            request.session['access_token'] = {'expires_at': datetime.utcnow() + timedelta(minutes=2)}
            request.session['user'] = {
                'id': self.userid,
                'login': 'x@x.x',
                'alias': 'XXXX XXXX',
                'main_tenant': 'XXXXXXXX',
                'tenants': [{'id': 'XXXXXXXX', 'name': 'Test'}],
                'site_id': 'XXXXXXXX',
                'locale': 'en',
                'timezone': 'utc',
            }
            return self.userid

    def effective_principals(self, _request):
        principals = [Everyone]
        if self.principals:
            principals += [Authenticated, 'user']
            principals += as_list(self.principals)
        return principals

    @staticmethod
    def remember(_request, _userid, **_kw):
        return []

    @staticmethod
    def forget(_request):
        return []


class FunctionalTestBase(unittest.TestCase):
    module = None
    root = None
    set_up_blob_dir = False
    principals = None
    dummy_redis = DummyRedis()

    def get_dummy_redis(self, _request, **_kwargs):
        return self.dummy_redis

    # noinspection PyPep8Naming
    def DummyRedisSessionFactory(self, secret, **_kwargs):
        return RedisSessionFactory(secret, client_callable=self.get_dummy_redis)

    def get_auth_patch(self):
        if self.principals:
            return patch(f'{self.module.__name__}.OpenIDConnectAuthenticationPolicy')
        else:
            return dummy_context_manager()

    # noinspection PyPep8Naming
    def setUp(self):
        config_file = self.root / 'config/tests.ini'
        global_config = {'__file__': config_file, 'here': self.root}

        parsed_settings = ConfigParser({'here': str(self.root)})
        parsed_settings.read(config_file)
        self.settings = dict(parsed_settings.items('app:main'))
        if self.set_up_blob_dir:
            self.blob_dir = tempfile.mkdtemp()
            self.settings[f'{self.module.__name__}.blobstore_path'] = self.blob_dir

        with self.get_auth_patch() as MockAuthenticationPolicy, \
                patch(f'{self.module.__name__}.RedisSessionFactory') as MockRedisSessionFactory:
            if self.principals:
                MockAuthenticationPolicy.side_effect = partial(DummyAuthenticationPolicy, principals=self.principals)
            MockRedisSessionFactory.side_effect = self.DummyRedisSessionFactory

            wsgi_app = self.module.main(global_config, **self.settings)

            self.session_factory = wsgi_app.registry['db_session_factory']
            self.engine = self.session_factory.kw['bind']

            alembic_cfg = alembic.config.Config(file_=str(config_file))
            alembic_cfg.attributes['engine'] = self.engine
            alembic_cfg.set_main_option('script_location', 'alembic')
            alembic.command.upgrade(alembic_cfg, 'head')

            extra_environ = {
                'REMOTE_ADDR': '127.0.0.1',
                'wsgi.url_scheme': 'https',
            }
            self.app = TestApp(wsgi_app, extra_environ=extra_environ)

    def tearDown(self):
        Model.metadata.drop_all(bind=self.engine)
        if self.set_up_blob_dir:
            shutil.rmtree(self.blob_dir)

    def dummy_request(self, user_is_admin=False):
        request = DummyRequest()
        request.user = User(is_admin=user_is_admin)
        request.db_session = self.session_factory()
        request.registry.settings = self.settings
        return request
