from functools import partialmethod


class AuthenticatedEndpoint(type):
    """This is a Pyramid view class metaclass. It has three objectives:
        - when an unauthenticated user requests a view, there is no need of calling the sometimes expensive acl and init
        class functions.
        - we can then write the acl and init functions with confidence that request.user exists.
        - finally, in our schema, with one class (and not one for the factory/acl and one for the view), the init
        function gets called twice. But the class instance from the first call is passed by pyramid as request.context
        in the second call. We can then reuse the instance __dict__ instead of calling the init a second time."""

    # noinspection PyMethodParameters
    def __new__(mcls, name, bases, namespace):
        """When creating the class, decorate its acl and init functions with our local ones."""
        cls = super().__new__(mcls, name, bases, namespace)
        cls.__acl__ = partialmethod(mcls.__auth_acl__, cls_acl=cls.__acl__)
        cls.__init__ = partialmethod(mcls.__auth_init__, cls_init=cls.__init__)
        return cls

    def __auth_acl__(self, cls_acl):
        if self.request.user and callable(cls_acl):
            return cls_acl(self)
        # cls.__acl__ can be a list.
        elif self.request.user:
            return cls_acl
        else:
            return []

    def __auth_init__(self, request, cls_init):
        self.request = request
        # During the first __init__() call, request.context does not exist.
        if request.user and not hasattr(request, 'context'):
            cls_init(self, request)
        # During the second __init__() call, reuse the __dict__ from the first call.
        elif request.user:
            self.__dict__ = request.context.__dict__
